import unittest
from talos.thrift.topic.ttypes import TopicTalosResourceName
from talos.thrift.topic.ttypes import Topic
from talos.thrift.topic.ttypes import GetDescribeInfoResponse
from talos.thrift.topic.ttypes import GetTopicAttributeResponse
from talos.thrift.topic.ttypes import TopicAttribute
from talos.thrift.message.ttypes import Message
from talos.thrift.message.ttypes import PutMessageResponse
from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.TalosClientFactory import TalosClientFactory
from talos.client.TalosClientFactory import MessageClient
from talos.client.TalosClientConfigkeys import TalosClientConfigKeys
from talos.client.SimpleAbnormalCallback import SimpleAbnormalCallback
from talos.admin.TalosAdmin import TalosAdmin
from talos.producer.TalosProducer import TalosProducer
from talos.producer.PartitionSender import PartitionSender
from talos.producer.UserMessageCallback import UserMessageCallback
from talos.producer.TalosProducerConfigKey import TalosProducerConfigKeys
from talos.utils.Utils import synchronized
from random import Random
from mock import Mock
import time

base = "abcdefgh ijklmnopqr stuvwxyz 0123456789"


# generate random string as message for putMessage
def get_random_string(self, randomStrLen=None):
	random = Random()
	stringBuffer = ""

	i = 0
	while i < randomStrLen:
		number = random.randint(0, len(base))
		stringBuffer += base[number]
		i += 1
	return stringBuffer


@synchronized
def add_success_counter(self, counter=None):
	self.msgPutSuccessCount += counter


@synchronized
def add_failure_counter(self, counter=None):
	self.msgPutFailureCount += counter


class TestCallback(UserMessageCallback):
	def on_success(self, userMessageResult=None):
		add_success_counter(counter=len(userMessageResult.get_message_list()))

	def on_error(self, userMessageResult=None):
		add_failure_counter(counter=len(userMessageResult.get_message_list()))


class test_TalosProducer(unittest.TestCase):
	resourceName = "12345#TopicName#july777777000999"
	anotherResourceName = "12345#TopicName#july777777000629"
	topicName = "TopicName"
	ownerId = "12345"
	messageRetentionMs = 1
	partitionNumber = 8
	partitionNumber2 = 16
	randomStrLen = 15
	producerMaxBufferedMillSecs = 0.01
	producerMaxPutMsgNumber = 10
	producerMaxPutMsgBytes = 100
	checkPartitionInterval = 0.2
	talosResourceName = TopicTalosResourceName(topicTalosResourceName=resourceName)

	talosProducerConfig = TalosClientConfig
	talosProducer = TalosProducer
	messageList = []
	topic = Topic

	talosAdminMock = TalosAdmin
	talosClientFactoryMock = TalosClientFactory
	messageClientMock = MessageClient
	partitionSenderMock = PartitionSender
	msgPutSuccessCount = int
	msgPutFailureCount = int

	def setUp(self):
		# set properties
		properties = dict()
		properties[TalosProducerConfigKeys.GALAXY_TALOS_PRODUCER_MAX_BUFFERED_MILLI_SECS] = \
			self.producerMaxBufferedMillSecs
		properties[TalosProducerConfigKeys.GALAXY_TALOS_PRODUCER_MAX_PUT_MESSAGE_NUMBER] = \
			self.producerMaxPutMsgNumber
		properties[TalosProducerConfigKeys.GALAXY_TALOS_PRODUCER_MAX_PUT_MESSAGE_BYTES] = \
			self.producerMaxPutMsgBytes
		properties[TalosProducerConfigKeys.GALAXY_TALOS_PRODUCER_CHECK_PARTITION_INTERVAL] = \
			self.checkPartitionInterval
		properties[TalosClientConfigKeys.GALAXY_TALOS_SERVICE_ENDPOINT] = "testURI"
		self.talosProducerConfig = TalosClientConfig(properties)

		# construct a getDescribeInfoResponse
		self.getDescribeInfoResponse = GetDescribeInfoResponse(TopicTalosResourceName(
			topicTalosResourceName=self.talosResourceName), self.partitionNumber)

		# construct a getTopicAttributeResponse
		self.getTopicAttributeResponse = GetTopicAttributeResponse(self.talosResourceName,
																   TopicAttribute(partitionNumber=self.partitionNumber))

		# mock some return value
		self.talosAdminMock = Mock(TalosAdmin)
		self.talosClientFactoryMock = Mock(TalosClientFactory)
		self.messageClientMock = Mock(MessageClient)
		self.partitionSenderMock = Mock(PartitionSender)

		# generate 100 random messages
		messageList = []
		i = 0
		while i < 100:
			messageList.append(Message(get_random_string(self.randomStrLen)))
			i += 1

		# mock putMessageResponse
		self.msgPutFailureCount = 0
		self.msgPutSuccessCount = 0

		self.talosClientFactoryMock.new_message_client = Mock(return_value=self.messageClientMock)
		self.messageClientMock.put_message = Mock(return_value=PutMessageResponse())
		self.talosAdminMock.get_describe_info = Mock(return_value=self.getDescribeInfoResponse)
		self.talosAdminMock.get_topic_attribute = Mock(return_value=self.getTopicAttributeResponse)

	def tearDown(self):
		pass

	def test_asynchronously_addUser_message(self):
		self.talosProducer = TalosProducer(producerConfig=self.talosProducerConfig,
										   topicName=self.topicName,
										   topicTalosResourceName=TopicTalosResourceName(topicTalosResourceName=self.talosResourceName),
										   talosAdmin=self.talosAdminMock,
										   talosClientFactory=self.talosClientFactoryMock,
										   topicAbnormalCallback=SimpleAbnormalCallback(),
										   userMessageCallback=TestCallback())

		self.talosProducer.add_user_message(self.messageList)
		# wait for execute finished
		time.sleep(self.producerMaxBufferedMillSecs * 10)

	def test_producer_not_active_error(self):
		getDescribeInfoResponse = GetDescribeInfoResponse(TopicTalosResourceName(
			TopicTalosResourceName(self.talosResourceName)), self.partitionNumber)
		getDescribeInfoResponse2 = GetDescribeInfoResponse(TopicTalosResourceName(
			TopicTalosResourceName(self.anotherResourceName)), self.partitionNumber)
		self.talosAdminMock.get_describe_info = Mock(side_effcet=[getDescribeInfoResponse, getDescribeInfoResponse2])

		self.partitionSenderMock.shutdown = Mock(return_value=None)

		self.talosProducer = TalosProducer(producerConfig=self.talosProducerConfig,
										   topicName=self.topicName,
										   topicTalosResourceName=TopicTalosResourceName(topicTalosResourceName=self.talosResourceName),
										   talosAdmin=self.talosAdminMock,
										   talosClientFactory=self.talosClientFactoryMock,
										   topicAbnormalCallback=SimpleAbnormalCallback(),
										   userMessageCallback=TestCallback())

		# wait check partition interval
		time.sleep(self.checkPartitionInterval * 2)

		self.partitionSenderMock.add_message = Mock(return_value=None)
		self.talosProducer.add_user_message(self.messageList)

	# check partition change when producer running
	def test_partition_change_during_producer_running(self):
		getDescribeInfoResponse = GetDescribeInfoResponse(TopicTalosResourceName(
			TopicTalosResourceName(self.talosResourceName)), self.partitionNumber)
		getDescribeInfoResponse2 = GetDescribeInfoResponse(TopicTalosResourceName(
			TopicTalosResourceName(self.anotherResourceName)), self.partitionNumber)
		self.talosAdminMock.get_describe_info = Mock(side_effect=[getDescribeInfoResponse,
																  getDescribeInfoResponse2])

		self.talosProducer = TalosProducer(producerConfig=self.talosProducerConfig,
										   topicName=self.topicName,
										   topicTalosResourceName=TopicTalosResourceName(topicTalosResourceName=self.talosResourceName),
										   talosAdmin=self.talosAdminMock,
										   talosClientFactory=self.talosClientFactoryMock,
										   topicAbnormalCallback=SimpleAbnormalCallback(),
										   userMessageCallback=TestCallback())

		# wait check partition interval
		time.sleep(self.checkPartitionInterval * 2)
		# check the partition number and outgoingMessageMap changing by log info


if __name__ == '__main__':
	unittest.main()



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 
 
import unittest

from talos.thrift.consumer.ttypes import LockWorkerResponse
from talos.thrift.consumer.ttypes import RenewResponse
from talos.thrift.consumer.ttypes import QueryWorkerResponse
from talos.thrift.topic.ttypes import TopicTalosResourceName
from talos.thrift.topic.ttypes import TopicInfo
from talos.thrift.topic.ttypes import TopicAttribute
from talos.thrift.topic.ttypes import Topic
from talos.thrift.topic.ttypes import TopicState
from talos.thrift.topic.ttypes import GetDescribeInfoResponse
from talos.thrift.common.ttypes import GalaxyTalosException
from talos.thrift.common.ttypes import ErrorCode
from talos.consumer.TalosConsumer import TalosConsumer
from talos.consumer.PartitionFetcher import PartitionFetcher
from talos.client.TalosClientFactory import ConsumerClient
from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.TalosClientConfigkeys import TalosClientConfigKeys
from talos.admin.TalosAdmin import TalosAdmin
from talos.client.TalosErrors import InvalidArgumentError
from talos.utils import Utils
from mock import Mock
import time


class test_TalosConsumer(unittest.TestCase):
    topicName = "MyTopic"
    ownerId = "12345"
    partitionNum = 5
    partitionNum2 = 6
    resourceName = TopicTalosResourceName(topicTalosResourceName="12345#MyTopic#34595fkdiso456i390")
    resourceName2 = TopicTalosResourceName(topicTalosResourceName="12345#MyTopic#34595fkdiso456i23")
    consumerGroup = "MyConsumerGroup"
    clientIdPrefix = "TestClient1"
    clientIdPrefix2 = "TestClient2"
    clientIdPrefix3 = "TestClient3"
    workerId = Utils.generate_client_id(clientIdPrefix, "")
    workerId2 = Utils.generate_client_id(clientIdPrefix2, "")
    workerId3 = Utils.generate_client_id(clientIdPrefix3, "")

    talosResourceName = resourceName
    talosResourceName2 = resourceName2
    topicInfo = TopicInfo(topicName, talosResourceName, ownerId)
    topicInfo2 = TopicInfo(topicName, talosResourceName2, ownerId)
    topicAttribute = TopicAttribute(partitionNumber=partitionNum)
    topicAttribute2 = TopicAttribute(partitionNumber=partitionNum2)
    topic = Topic(topicInfo, topicAttribute, TopicState())
    # resourceName changed
    topic2 = Topic(topicInfo2, topicAttribute, TopicState())
    # partitionNumber changed
    topic3 = Topic(topicInfo, topicAttribute2, TopicState())

    fetcherMap = dict()
    partitionFetcher0 = Mock(PartitionFetcher)
    partitionFetcher1 = Mock(PartitionFetcher)
    partitionFetcher2 = Mock(PartitionFetcher)
    partitionFetcher3 = Mock(PartitionFetcher)
    partitionFetcher4 = Mock(PartitionFetcher)
    partitionFetcher5 = Mock(PartitionFetcher)

    talosAdminMock = Mock(TalosAdmin)
    consumerClientMock = Mock(ConsumerClient)

    def setUp(self):
        properties = dict()
        # check partition number interval 200ms
        properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_CHECK_PARTITION_INTERVAL] = 0.2
        # check worker info interval 300ms
        properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_CHECK_WORKER_INFO_INTERVAL] = 0.3
        # renew interval 300ms
        properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_RENEW_INTERVAL] = 0.3
        properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_RENEW_MAX_RETRY] = 0
        properties[TalosClientConfigKeys.GALAXY_TALOS_CONSUMER_REGISTER_MAX_RETRY] = 0
        properties[TalosClientConfigKeys.GALAXY_TALOS_SERVICE_ENDPOINT] = "testURI"

        # do not check parameter validity for unit test
        self.consumerConfig = TalosClientConfig(properties)

        self.fetcherMap[0] = self.partitionFetcher0
        self.fetcherMap[1] = self.partitionFetcher1
        self.fetcherMap[2] = self.partitionFetcher2
        self.fetcherMap[3] = self.partitionFetcher3
        self.fetcherMap[4] = self.partitionFetcher4
        # self.fetcherMap[5] = self.partitionFetcher5

        # construct a getDescribeInfoResponse
        getDescribeInfoResponse = GetDescribeInfoResponse(self.talosResourceName, self.partitionNum)
        self.talosAdminMock.get_describe_info = Mock(return_value=getDescribeInfoResponse)

    def tearDown(self):
        pass

    # 1) checkAndGetTopicInfoFailed
    def test_CheckAndGetTopicInfoFailed(self):
        # get describe info
        getDescribeInfoResponse = GetDescribeInfoResponse(self.talosResourceName2, self.partitionNum)
        self.talosAdminMock.get_describe_info = Mock(return_value=getDescribeInfoResponse)

        with self.assertRaises(InvalidArgumentError) as context:
            TalosConsumer(consumerGroup=self.consumerGroup, consumerConfig=self.consumerConfig,
                          talosResourceName=TopicTalosResourceName(self.talosResourceName),
                          workerId=self.workerId, consumerClient=self.consumerClientMock,
                          talosAdmin=self.talosAdminMock, fetcherMap=self.fetcherMap)

    # 2) registerSelf failed, init failed
    def test_RegisterSelfFailed(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(False)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)

        with self.assertRaises(RuntimeError) as context:
            TalosConsumer(consumerGroup=self.consumerGroup, consumerConfig=self.consumerConfig,
                          talosResourceName=TopicTalosResourceName(self.talosResourceName),
                          workerId=self.workerId, consumerClient=self.consumerClientMock,
                          talosAdmin=self.talosAdminMock, fetcherMap=self.fetcherMap)

    # 3) lock -> balance
    def test_LockBalance(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        # worker info check
        workerInfo = dict()
        partitionList = [3, 0, 2]
        workerInfo[self.workerId] = []
        workerInfo[self.workerId2] = partitionList
        workerInfo2 = dict()
        partitionList2 = [1, 4]
        workerInfo2[self.workerId] = partitionList2
        workerInfo2[self.workerId2] = partitionList
        # first not balance, then balance for serving partition number
        self.consumerClientMock.query_worker = Mock(side_effect=[QueryWorkerResponse(workerInfo),
                                                                 QueryWorkerResponse(workerInfo2)])
        # for renew/steal partition/re-balance
        self.partitionFetcher0.is_serving = Mock(side_effect=[False, False])
        self.partitionFetcher1.is_serving = Mock(side_effect=[False, True])
        self.partitionFetcher2.is_serving = Mock(side_effect=[False, False])
        self.partitionFetcher3.is_serving = Mock(side_effect=[False, False])
        self.partitionFetcher4.is_serving = Mock(side_effect=[False, True])

        self.partitionFetcher0.is_holding_lock = Mock(side_effect=[False, False])
        self.partitionFetcher1.is_holding_lock = Mock(side_effect=[False, True])
        self.partitionFetcher2.is_holding_lock = Mock(side_effect=[False, False])
        self.partitionFetcher3.is_holding_lock = Mock(side_effect=[False, False])
        self.partitionFetcher4.is_holding_lock = Mock(side_effect=[False, True])

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)
        # sleep for 2 times balance
        time.sleep(0.8)
        consumer.shutdown()

    # 4) worker online: lock -> unlock -> balance
    def test_WorkerOnlineLockUnlock(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        # worker info check
        # first: 1 worker, 5, this worker lock 5 partitions
        workerInfo = dict()
        workerInfo[self.workerId] = []
        # second: 3 worker, 2 2 1, has sort: 5 0 0, this worker unlock 3 partition
        workerInfo2 = dict()
        workerInfo2[self.workerId] = [0, 1, 2, 3, 4]
        workerInfo2[self.workerId2] = []
        workerInfo2[self.workerId3] = []
        self.consumerClientMock.query_worker = Mock(side_effect=[QueryWorkerResponse(workerInfo),
                                                                 QueryWorkerResponse(workerInfo2)])
        # for renew/steal partition/re-balance
        # construct 3 times balance
        self.partitionFetcher0.is_serving = Mock(side_effect=[False, True, False])
        self.partitionFetcher1.is_serving = Mock(side_effect=[False, True, False])
        self.partitionFetcher2.is_serving = Mock(side_effect=[False, True, False])
        self.partitionFetcher3.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher4.is_serving = Mock(side_effect=[False, True, True])

        self.partitionFetcher0.is_holding_lock = Mock(side_effect=[False, True, False])
        self.partitionFetcher1.is_holding_lock = Mock(side_effect=[False, True, False])
        self.partitionFetcher2.is_holding_lock = Mock(side_effect=[False, True, False])
        self.partitionFetcher3.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher4.is_holding_lock = Mock(side_effect=[False, True, True])

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 3 times balance
        time.sleep(1)
        consumer.shutdown()

    # 5) worker offline: lock -> lock -> balance
    def test_WorkerOfflineLockLock(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        # worker info check
        # first: 2 worker, [3, 2], this worker lock 2 paritions
        workerInfo = dict()
        partitionList = [4, 0, 2]
        workerInfo[self.workerId] = []
        workerInfo[self.workerId2] = partitionList
        # second: 1 worker, the other worker die, this worker lock 3 paritions
        workerInfo2 = dict()
        partitionList2 = [3, 1]
        workerInfo2[self.workerId] = partitionList2
        self.consumerClientMock.query_worker = Mock(side_effect=[QueryWorkerResponse(workerInfo),
                                                                 QueryWorkerResponse(workerInfo2)])
        # for renew / steal partition / re-balance
        # construct 2 times balance
        self.partitionFetcher1.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher3.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher0.is_serving = Mock(side_effect=[False, False, True])
        self.partitionFetcher2.is_serving = Mock(side_effect=[False, False, True])
        self.partitionFetcher4.is_serving = Mock(side_effect=[False, False, True])

        self.partitionFetcher1.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher3.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher0.is_holding_lock = Mock(side_effect=[False, False, True])
        self.partitionFetcher2.is_holding_lock = Mock(side_effect=[False, False, True])
        self.partitionFetcher4.is_holding_lock = Mock(side_effect=[False, False, True])

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 3 times balance
        time.sleep(1)
        consumer.shutdown()

    # 6) partition increase: lock -> lock -> balance
    def test_PartitionChangedLockLock(self):
        self.fetcherMap[5] = self.partitionFetcher5
        # partition check, change from 5 to 6
        getDescribeInfoResponse = GetDescribeInfoResponse(self.talosResourceName, self.partitionNum)
        getDescribeInfoResponse3 = GetDescribeInfoResponse(self.talosResourceName, self.partitionNum2)
        self.talosAdminMock.get_describe_info = Mock(side_effect=[getDescribeInfoResponse, getDescribeInfoResponse3])
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        # worker info check
        # first: 1 worker 5 partitions
        workerInfo = dict()
        workerInfo[self.workerId] = []
        # second: 1 worker 6 partitions
        workerInfo2 = dict()
        partitionList = [0, 1, 2, 3, 4]
        workerInfo2[self.workerId] = partitionList
        self.consumerClientMock.query_worker = Mock(side_effect=[QueryWorkerResponse(workerInfo),
                                                                 QueryWorkerResponse(workerInfo2),
                                                                 QueryWorkerResponse(workerInfo2)])
        # for renew / steal partition / re-balance
        # construct 2 times balance
        self.partitionFetcher0.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher1.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher2.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher3.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher4.is_serving = Mock(side_effect=[False, True, True])
        self.partitionFetcher5.is_serving = Mock(side_effect=[False, False, True])

        self.partitionFetcher0.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher1.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher2.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher3.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher4.is_holding_lock = Mock(side_effect=[False, True, True])
        self.partitionFetcher5.is_holding_lock = Mock(side_effect=[False, False, True])

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 3 times balance
        time.sleep(1)
        consumer.shutdown()

    # 7) renew failed: lock -> unlock -> balance
    def test_RenewFailedLockUnlock(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        failedList = [0]
        self.consumerClientMock.renew = Mock(side_effect=[RenewResponse(True, []), RenewResponse(True, failedList)])
        # worker info check
        # first: 1 worker 5 partitions
        workerInfo = dict()
        partitionList = [0, 1, 2, 3, 4]
        workerInfo[self.workerId] = partitionList
        self.consumerClientMock.query_worker = Mock(return_value=QueryWorkerResponse(workerInfo))
        # for renew / steal partition / re-balance
        self.partitionFetcher0.is_serving = Mock(return_value=True)
        self.partitionFetcher1.is_serving = Mock(return_value=True)
        self.partitionFetcher2.is_serving = Mock(return_value=True)
        self.partitionFetcher3.is_serving = Mock(return_value=True)
        self.partitionFetcher4.is_serving = Mock(return_value=True)

        self.partitionFetcher0.is_holding_lock = Mock(side_effect=[True, False])
        self.partitionFetcher1.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher2.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher3.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher4.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher4.is_holding_lock = Mock(return_value=False)

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 3 times renew
        time.sleep(0.8)
        consumer.shutdown()

    # 8) renew heartbeat failed: cancel all task
    def test_RenewHeartbeatFailed(self):
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(False, []))
        # worker info check
        # first: 1 worker 5 partitions, then renew failed
        workerInfo = dict()
        workerInfo[self.workerId] = [0, 1, 2, 3, 4]
        self.consumerClientMock.query_worker = Mock(return_value=QueryWorkerResponse(workerInfo))
        # for renew / steal partition / re-balance
        self.partitionFetcher0.is_serving = Mock(return_value=True)
        self.partitionFetcher1.is_serving = Mock(return_value=True)
        self.partitionFetcher2.is_serving = Mock(return_value=True)
        self.partitionFetcher3.is_serving = Mock(return_value=True)
        self.partitionFetcher4.is_serving = Mock(return_value=True)

        self.partitionFetcher0.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher1.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher2.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher3.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher4.is_holding_lock = Mock(return_value=True)

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 3 times renew
        time.sleep(0.8)
        consumer.shutdown()

    # 9) describeTopic not exist error
    def test_DescribeTopicNotExist(self):
        getDescribeInfoResponse = GetDescribeInfoResponse(self.talosResourceName, self.partitionNum)
        self.talosAdminMock.get_describe_info = Mock(side_effect=[getDescribeInfoResponse,
                                                                  GalaxyTalosException(errorCode=ErrorCode.TOPIC_NOT_EXIST)])
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        workerInfo = dict()
        workerInfo[self.workerId] = []
        self.consumerClientMock.query_worker = Mock(return_value=QueryWorkerResponse(workerInfo))

        self.partitionFetcher0.is_serving = Mock(return_value=False)
        self.partitionFetcher1.is_serving = Mock(return_value=False)
        self.partitionFetcher2.is_serving = Mock(return_value=False)
        self.partitionFetcher3.is_serving = Mock(return_value=False)
        self.partitionFetcher4.is_serving = Mock(return_value=False)

        self.partitionFetcher0.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher1.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher2.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher3.is_holding_lock = Mock(return_value=True)
        self.partitionFetcher4.is_holding_lock = Mock(return_value=True)

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 1 times renew
        time.sleep(0.31)
        consumer.shutdown()

    # 10) describeTopic resourceName changed, old topic deleted
    def test_DescribeTopicResourceNameChanged(self):
        getDescribeInfoResponse = GetDescribeInfoResponse(self.talosResourceName, self.partitionNum)
        getDescribeInfoResponse2 = GetDescribeInfoResponse(self.talosResourceName2, self.partitionNum)
        self.talosAdminMock.get_describe_info = Mock(side_effect=[getDescribeInfoResponse, getDescribeInfoResponse2])
        # register self
        lockWorkerResponse = LockWorkerResponse(True)
        self.consumerClientMock.lock_worker = Mock(return_value=lockWorkerResponse)
        # renew check
        self.consumerClientMock.renew = Mock(return_value=RenewResponse(True, []))
        workerInfo = dict()
        workerInfo[self.workerId] = []
        self.consumerClientMock.query_worker = Mock(return_value=QueryWorkerResponse(workerInfo))

        self.partitionFetcher0.is_serving = Mock(return_value=False)
        self.partitionFetcher1.is_serving = Mock(return_value=False)
        self.partitionFetcher2.is_serving = Mock(return_value=False)
        self.partitionFetcher3.is_serving = Mock(return_value=False)
        self.partitionFetcher4.is_serving = Mock(return_value=False)

        consumer = TalosConsumer(consumerGroup=self.consumerGroup,
                                 consumerConfig=self.consumerConfig,
                                 talosResourceName=TopicTalosResourceName(self.talosResourceName),
                                 workerId=self.workerId,
                                 consumerClient=self.consumerClientMock,
                                 talosAdmin=self.talosAdminMock,
                                 fetcherMap=self.fetcherMap)

        # sleep for 1 times renew
        time.sleep(0.4)
        consumer.shutdown()






#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.utils.Utils import synchronized
from talos.utils import Utils
from talos.thrift.common.ttypes import GalaxyTalosException
import logging
import threading
import traceback


class PartitionMessageQueue:
    logging.basicConfig()
    logger = logging.getLogger("PartitionMessageQueue")
    userMessageList = []
    curMessageBytes = int
    partitionId = int
    maxBufferedTime = int
    maxPutMsgNumber = int
    maxPutMsgBytes = int

    def __init__(self, producerConfig=None, partitionId=None, producerPtr=None):
        self.userMessageList = []
        self.curMessageBytes = 0
        self.partitionId = partitionId
        self.producer = producerPtr
        self.maxBufferedTime = producerConfig.get_max_buffered_msg_time()
        self.maxPutMsgNumber = producerConfig.get_max_put_msg_number()
        self.maxPutMsgBytes = producerConfig.get_max_put_msg_bytes()
        self.globalLock = threading.Condition()
        self.synchronizedLock = threading.Lock()

    def add_message(self, messageList=None):
        if self.globalLock.acquire():
            incrementBytes = 0
            for userMessage in messageList:
                self.userMessageList.insert(0, userMessage)
                incrementBytes += userMessage.get_message_size()

            self.curMessageBytes += incrementBytes
            # update total buffered count when add messageList
            self.producer.increase_buffered_count(len(messageList), incrementBytes)

            # notify partitionSender to getUserMessageList
            self.globalLock.notifyAll()
            self.globalLock.release()

    #
    # return messageList, if not shouldPut, block in this method
    #
    def get_message_list(self):
        if self.globalLock.acquire():
            while not self.should_put():
                try:
                    waitTime = self.get_wait_time()
                    self.globalLock.wait(waitTime)
                except Exception as e:
                    if isinstance(e, GalaxyTalosException):
                        self.logger.error(e.errMsg)
                    self.logger.error("getUserMessageList for partition: " +
                                      str(self.partitionId) + " is interrupt when waiting: "
                                      + traceback.format_exc())

        if self.logger.isEnabledFor(level=logging.DEBUG):
            self.logger.debug("getUserMessageList wake up for partition: "
                              + str(self.partitionId))

        returnList = []
        returnMsgBytes = 0
        returnMsgNumber = 0

        while (not len(self.userMessageList) == 0) and \
                returnMsgNumber < self.maxPutMsgNumber and \
                returnMsgBytes < self.maxPutMsgBytes:
            userMessage = self.userMessageList[len(self.userMessageList) - 1]
            del self.userMessageList[len(self.userMessageList) - 1]
            returnList.append(userMessage.get_message())
            self.curMessageBytes -= userMessage.get_message_size()
            returnMsgBytes += userMessage.get_message_size()
            returnMsgNumber += 1

        # update total buffered count when poll messageList
        self.producer.decrease_buffered_count(returnMsgNumber, returnMsgBytes)
        if self.logger.isEnabledFor(level=logging.DEBUG):
            self.logger.debug("Ready to put message batch: " + str(len(returnList)) +
                              " queue size: " + str(len(self.userMessageList))
                              + " and curBytes: " + str(self.curMessageBytes) +
                              " for partition: " + str(len(self.partitionId)))
        return returnList

    @synchronized
    def should_put(self):
        # when TalosProducer is not active;
        if not self.producer.is_active():
            return True

        # when we have enough bytes data or enough number data;
        if self.curMessageBytes >= self.maxPutMsgBytes or \
                len(self.userMessageList) >= self.maxPutMsgNumber:
            return True

        # when there have at least one message and it has exist enough long time;
        if len(self.userMessageList) > 0:
            curBufferedTime = Utils.current_time_mills() - self.userMessageList[len(self.userMessageList) - 1].get_timestamp()
            if curBufferedTime >= self.maxBufferedTime:
                return True

        return False

    #
    # Note: wait(0) represents wait infinite until be notified
    # so we wait minimal 1 milli secs when time <= 0
    #
    @synchronized
    def get_wait_time(self):
        if len(self.userMessageList) <= 0:
            return 0
        time = (self.userMessageList[len(self.userMessageList) - 1].get_timestamp()
                + self.maxBufferedTime - Utils.current_time_mills()) / 1000
        if time > 0:
            return time
        else:
            return 1

    #
    # clean up MessageQueue, return all message list
    #
    @synchronized
    def get_all_message_list(self):
        returnList = []
        returnMsgBytes = 0
        returnMsgNumber = 0

        while not len(self.userMessageList) == 0:
            userMessage = self.userMessageList[len(self.userMessageList) - 1]
            del self.userMessageList[len(self.userMessageList) - 1]
            returnList.append(userMessage.get_message())
            self.curMessageBytes -= userMessage.get_message_size()
            returnMsgBytes += userMessage.get_message_size()
            returnMsgNumber += 1

        # update total buffered count when poll messageList
        self.producer.decrease_buffered_count(returnMsgNumber, returnMsgBytes)
        return returnList

    @synchronized
    def get_cur_message_bytes(self):
        return self.curMessageBytes











#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

import logging
from talos.thrift.common.ttypes import *
from talos.thrift.Thrift import TProcessor
from talos.thrift.transport import TTransport
from talos.thrift.protocol import TBinaryProtocol, TProtocol
try:
  from talos.thrift.protocol import fastbinary
except:
  fastbinary = None


class Iface:
  def getServiceVersion(self):
    """
    Get Talas service version;

    """
    pass

  def validClientVersion(self, clientVersion):
    """
    Check the version compatibility between client and server;


    Parameters:
     - clientVersion
    """
    pass


class Client(Iface):
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def getServiceVersion(self):
    """
    Get Talas service version;

    """
    self.send_getServiceVersion()
    return self.recv_getServiceVersion()

  def send_getServiceVersion(self):
    self._oprot.writeMessageBegin('getServiceVersion', TMessageType.CALL, self._seqid)
    args = getServiceVersion_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getServiceVersion(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getServiceVersion_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.e is not None:
      raise result.e
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getServiceVersion failed: unknown result")

  def validClientVersion(self, clientVersion):
    """
    Check the version compatibility between client and server;


    Parameters:
     - clientVersion
    """
    self.send_validClientVersion(clientVersion)
    self.recv_validClientVersion()

  def send_validClientVersion(self, clientVersion):
    self._oprot.writeMessageBegin('validClientVersion', TMessageType.CALL, self._seqid)
    args = validClientVersion_args()
    args.clientVersion = clientVersion
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_validClientVersion(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = validClientVersion_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.e is not None:
      raise result.e
    return


class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["getServiceVersion"] = Processor.process_getServiceVersion
    self._processMap["validClientVersion"] = Processor.process_validClientVersion

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_getServiceVersion(self, seqid, iprot, oprot):
    args = getServiceVersion_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getServiceVersion_result()
    try:
      result.success = self._handler.getServiceVersion()
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except GalaxyTalosException as e:
      msg_type = TMessageType.REPLY
      result.e = e
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getServiceVersion", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_validClientVersion(self, seqid, iprot, oprot):
    args = validClientVersion_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = validClientVersion_result()
    try:
      self._handler.validClientVersion(args.clientVersion)
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except GalaxyTalosException as e:
      msg_type = TMessageType.REPLY
      result.e = e
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("validClientVersion", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class getServiceVersion_args:

  thrift_spec = (
  )

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getServiceVersion_args')
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getServiceVersion_result:
  """
  Attributes:
   - success
   - e
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (Version, Version.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'e', (GalaxyTalosException, GalaxyTalosException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, e=None,):
    self.success = success
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = Version()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.e = GalaxyTalosException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getServiceVersion_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.e)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class validClientVersion_args:
  """
  Attributes:
   - clientVersion
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'clientVersion', (Version, Version.thrift_spec), None, ), # 1
  )

  def __init__(self, clientVersion=None,):
    self.clientVersion = clientVersion

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.clientVersion = Version()
          self.clientVersion.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('validClientVersion_args')
    if self.clientVersion is not None:
      oprot.writeFieldBegin('clientVersion', TType.STRUCT, 1)
      self.clientVersion.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.clientVersion)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class validClientVersion_result:
  """
  Attributes:
   - e
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'e', (GalaxyTalosException, GalaxyTalosException.thrift_spec), None, ), # 1
  )

  def __init__(self, e=None,):
    self.e = e

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.e = GalaxyTalosException()
          self.e.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('validClientVersion_result')
    if self.e is not None:
      oprot.writeFieldBegin('e', TType.STRUCT, 1)
      self.e.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.e)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.items()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 


class TalosClientConfigKeys:

    # The client max retry times before throw exception
    GALAXY_TALOS_CLIENT_MAX_RETRY = "galaxy.talos.client.max.retry"
    GALAXY_TALOS_CLIENT_MAX_RETRY_DEFAULT = 1

    # The client timeout milli secs when write / read
    GALAXY_TALOS_CLIENT_TIMEOUT_MILLI_SECS = "galaxy.talos.client.timeout.milli.secs"
    GALAXY_TALOS_CLIENT_TIMEOUT_MILLI_SECS_DEFAULT = 10000

    # The client connection timeout
    GALAXY_TALOS_CLIENT_CONN_TIMECOUT_MILLI_SECS = "galaxy.talos.client.conn.timeout.milli.secs"
    GALAXY_TALOS_CLIENT_CONN_TIMECOUT_MILLI_SECS_DEFAULT = 5000

    # The client DDL operation timeout
    GALAXY_TALOS_CLIENT_ADMIN_TIMEOUT_MILLI_SECS = "galaxy.talos.client.admin.timeout.milli.secs"
    GALAXY_TALOS_CLIENT_ADMIN_TIMEOUT_MILLI_SECS_DEFAULT = 30000

    # Constants for http / https rpc address
    GALAXY_TALOS_SERVICE_ENDPOINT = "galaxy.talos.service.endpoint"
    GALAXY_TALOS_DEFAULT_SERVICE_ENDPOINT = "http://talos.api.xiaomi.com"

    GALAXY_TALOS_SECURE_SERVICE_ENDPOINT = "galaxy.talos.secure.service.endpoint"
    GALAXY_TALOS_DEFAULT_SECURE_SERVICE_ENDPOINT = "https://talos.api.xiaomi.com"

    # The http client connection params
    GALAXY_TALOS_HTTP_MAX_TOTAL_CONNECTION = "galaxy.talos.http.max.total.connection"
    GALAXY_TALOS_HTTP_MAX_TOTAL_CONNECTION_DEFAULT = 160

    GALAXY_TALOS_HTTP_MAX_TOTAL_CONNECTION_PER_ROUTE = "galaxy.talos.http.max.total.connection.per.route"
    GALAXY_TALOS_HTTP_MAX_TOTAL_CONNECTION_PER_ROUTE_DEFAULT = 160

    # The client whether to retry
    GALAXY_TALOS_CLIENT_IS_RETRY = "galaxy.talos.client.is.retry"
    GALAXY_TALOS_CLIENT_IS_RETRY_DEFAULT = False

    # The client whether auto location the serving restserver
    GALAXY_TALOS_CLIENT_IS_AUTO_LOCATION = "galaxy.talos.client.is.auto.location"
    GALAXY_TALOS_CLIENT_IS_AUTO_LOCATION_DEFAULT = True

    # The client get schedule info max retry
    GALAXY_TALOS_CLIENT_SCHEDULE_INFO_MAX_RETRY = "galaxy.talos.client.schedule.info.max.retry"
    GALAXY_TALOS_CLIENT_SCHEDULE_INFO_MAX_RETRY_DEFAULT = 1

    # interval for client update it's scheduleinfo cache
    GALSXY_TALOS_CLIENT_SCHEDULE_INFO_INTERVAL = "galaxy.talos.client.schedule.info.interval"
    GALAXY_TALOS_CLIENT_SCHEDULE_INFO_INTERVAL_DEFAULT = 1000 * 60 * 10

    # Talos consumer config
    # The consumer scan / update partition number interval(milli secs)
    GALAXY_TALOS_CONSUMER_CHECK_PARTITION_INTERVAL = "galaxy.talos.consumer.check.partition.interval"
    GALAXY_TALOS_CONSUMER_CHECK_PARTITION_INTERVAL_DEFAULT = 60
    GALAXY_TALOS_CONSUMER_CHECK_PARTITION_INTERVAL_MINIMUM = 60
    GALAXY_TALOS_CONSUMER_CHECK_PARTITION_INTERVAL_MAXIMUM = 60

    # The consumer check alive worker info and their serving partitions interval
    GALAXY_TALOS_CONSUMER_CHECK_WORKER_INFO_INTERVAL = "galaxy.talos.consumer.check.worker.info.interval"
    GALAXY_TALOS_CONSUMER_CHECK_WORKER_INFO_INTERVAL_DEFAULT = 10
    GALAXY_TALOS_CONSUMER_CHECK_WORKER_INFO_INTERVAL_MINIMUM = 10
    GALAXY_TALOS_CONSUMER_CHECK_WORKER_INFO_INTERVAL_MAXIMUM = 30

    # The consumer renew interval for both heartbeat and renew serving partitions
    # the worker column family ttl is 30s by default
    GALAXY_TALOS_CONSUMER_RENEW_INTERVAL = "galaxy.talos.consumer.renew.interval"
    GALAXY_TALOS_CONSUMER_RENEW_INTERVAL_DEFAULT = 7
    GALAXY_TALOS_CONSUMER_RENEW_INTERVAL_MINIMUM = 7
    GALAXY_TALOS_CONSUMER_RENEW_INTERVAL_MAXIMUM = 10

    # The consumer renew max retry
    GALAXY_TALOS_CONSUMER_RENEW_MAX_RETRY = "galaxy.talos.consumer.renew.max.retry"
    GALAXY_TALOS_CONSUMER_RENEW_MAX_RETRY_DEFAULT = 1
    GALAXY_TALOS_CONSUMER_RENEW_MAX_RETRY_MINIMUM = 1
    GALAXY_TALOS_CONSUMER_RENEW_MAX_RETRY_MAXIMUM = 3

    # The consumer getRecords max fetch message number
    GALAXY_TALOS_CONSUMER_MAX_FETCH_RECORDS = "galaxy.talos.consumer.max.fetch.records"
    GALAXY_TALOS_CONSUMER_MAX_FETCH_RECORDS_DEFAULT = 1000
    GALAXY_TALOS_CONSUMER_MAX_FETCH_RECORDS_MINIMUM = 1
    GALAXY_TALOS_CONSUMER_MAX_FETCH_RECORDS_MAXIMUM = 2000

    # The consumer getRecords max fetch message bytes
    GALAXY_TALOS_CONSUMER_MAX_FETCH_BYTES = "galaxy.talos.consumer.max.fetch.bytes"
    GALAXY_TALOS_CONSUMER_MAX_FETCH_BYTES_DEFAULT = 2 * 1024 * 1024
    GALAXY_TALOS_CONSUMER_MAX_FETCH_BYTES_MINIMUM = 1
    GALAXY_TALOS_CONSUMER_MAX_FETCH_BYTES_MAXIMUM = 10 * 1024 * 1024

    # The consumer fetch message operation interval
    # Note server GetRecords qps is [1, 5], the minimal interval is 200ms
    GALAXY_TALOS_CONSUMER_FETCH_INTERVAL = "galaxy.talos.consumer.fetch.interval.ms"
    GALAXY_TALOS_CONSUMER_FETCH_INTERVAL_DEFAULT = 200
    GALAXY_TALOS_CONSUMER_FETCH_INTERVAL_MINIMUM = 50
    GALAXY_TALOS_CONSUMER_FETCH_INTERVAL_MAXIMUM = 800

    # The consumer worker register self max retry times
    GALAXY_TALOS_CONSUMER_REGISTER_MAX_RETRY = "galaxy.talos.consumer.register.max.retry"
    GALAXY_TALOS_CONSUMER_REGISTER_MAX_RETRY_DEFAULT = 1

    # The consumer commit offset fetched records number threshold
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_THRESHOLD = "galaxy.talos.consumer.commit.offset.record.fetched.num"
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_THRESHOLD_DEFAULT = 10000
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_THRESHOLD_MINIMUM = 5000
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_THRESHOLD_MAXIMUM = 20000

    # The consumer commit offset time interval threshold, milli secs
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_INTERVAL = "galaxy.talos.consumer.commit.offset.interval.milli"
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_INTERVAL_DEFAULT = 5
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_INTERVAL_MINIMUM = 3
    GALAXY_TALOS_CONSUMER_COMMIT_OFFSET_INTERVAL_MAXIMUM = 8

    # The consumer switch for whether checking lastCommitOffset or not
    # when commit offset
    GALAXY_TALOS_CONSUMER_CHECK_LAST_COMMIT_OFFSET_SWITCH = "galaxy.talos.consumer.check.last.commit.offset.switch"
    GALAXY_TALOS_CONSUMER_CHECK_LAST_COMMIT_OFFSET_SWITCH_DEFAULT = False

    # The consumer partitionFetcher sleep / delay time when partitionNotServing
    GALAXY_TALOS_CONSUMER_WAIT_PARTITION_WORKING_TIME = "galaxy.talos.consumer.wait.partition.working.time.milli"
    GALAXY_TALOS_CONSUMER_WAIT_PARTITION_WORKING_TIME_DEFAULT = 0.2

    # The consumer reset offset by the latest offset when out of range
    GALAXY_TALOS_CONSUMER_OUT_OF_RANGE_RESET_LATEST_OFFSET = "galaxy.talos.consumer.out.of.range.reset.latest.offset"
    GALAXY_TALOS_CONSUMER_OUT_OF_RANGE_RESET_LATEST_OFFSET_DEFAULT = False

    # The consumer checkpoint auto commit
    GALAXY_TALOS_CONSUMER_CHECKPOINT_AUTO_COMMIT = "galaxy.talos.consumer.checkpoint.auto.commit"
    GALAXY_TALOS_CONSUMER_CHECKPOINT_AUTO_COMMIT_DEFAULT = True

    # When the consumer starts up(including firststart and restart),
    # this configuration indicates whether reset the offset which from reading.
    # The default value is false which means reading messages from the
    # 'last commit offset'(restart) or 'MessageOffset.START_OFFSET'(firstly start).
    GALAXY_TALOS_CONSUMER_START_WHETHER_RESET_OFFSET = "galaxy.talos.consumer.start.whether.reset.offset"
    GALAXY_TALOS_CONSUMER_START_WHETHER_RESET_OFFSET_DEFAULT = False

    # In the following conditions, this configuration will be effective:
    # 1) 'galaxy.talos.consumer.restart.whether.reset.offset' is 'true';
    # 2) the consumer is firstly starting which means there is not historical committed offset
    # Either condition above will lead to the consumer reading messages from the 'reset offset'.
    # The value of 'reset offset' has two kinds: -1, -2.
    # '-1' represents reading message from 'MessageOffset.START_OFFSET'
    # '-2' represents reading message from 'MessageOffset.LATEST_OFFSET'
    GALAXY_TALOS_CONSUMER_START_RESET_OFFSET_VALUE = "galaxy.talos.consumer.start.reset.offset.value"
    GALAXY_TALOS_CONSUMER_START_RESET_OFFSET_AS_START = -1
    GALAXY_TALOS_CONSUMER_START_RESET_OFFSET_AS_END = -2



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.utils.Utils import synchronized
import threading


class BufferedMessageCount:

    def __init__(self, maxBufferedMsgNumber, maxBufferedMsgBytes):
        self._maxBufferedMsgNumber = maxBufferedMsgNumber
        self._maxBufferedMsgBytes = maxBufferedMsgBytes
        self._bufferedMsgNumber = 0
        self._bufferedMsgBytes = 0
        self.synchronizedLock = threading.Lock()

    def get_buffered_msg_number(self):
        return self._bufferedMsgNumber

    def get_buffered_msg_bytes(self):
        return self._bufferedMsgBytes

    @synchronized
    def increase(self, diffBufferedMsgNumber, diffBufferedMsgBytes):
        self._bufferedMsgNumber += diffBufferedMsgNumber
        self._bufferedMsgBytes += diffBufferedMsgBytes

    @synchronized
    def decrease(self, diffBufferedMsgNumber, diffBufferedMsgBytes):
        self._bufferedMsgNumber -= diffBufferedMsgNumber
        self._bufferedMsgBytes -= diffBufferedMsgBytes

    @synchronized
    def is_empty(self):
        return self._bufferedMsgNumber == 0

    @synchronized
    def is_full(self):
        return self._bufferedMsgNumber >= self._maxBufferedMsgNumber or \
               self._bufferedMsgBytes >= self._maxBufferedMsgBytes



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

import abc
from talos.client.serialization.MessageVersion import MessageVersion
from talos.utils.Utils import SerializeFormat
from struct import pack, unpack


class MessageSerializer(object):

    VERSION_NUMBER_LENGTH = 4
    _CHARSET = "utf-8"

    @classmethod
    def decode_message_version(cls, header=None):
        if chr(header[0]) != 'V':
            return MessageVersion(1)
        else:
            versionNumberTuple = unpack(SerializeFormat.format_i16, header[1:3])
            versionNumber = versionNumberTuple[0]
            return MessageVersion(versionNumber)

    @classmethod
    def write_message_version(cls, version=None, buffer=None):
        versionNumber = version.get_version_number()
        buffer.write(b'V')
        buffer.write(pack(SerializeFormat.format_i16, versionNumber))
        buffer.write(pack('b', 0))

    # TalosProducer serialize
    @abc.abstractmethod
    def serialize(self, msg=None, buf=None):
        pass

    # TalosConsumer deserialize
    @abc.abstractmethod
    def deserialize(self, header=None, buf=None):
        pass

    @abc.abstractmethod
    def get_message_size(self, msg=None):
        pass










#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.producer.UserMessageCallback import UserMessageCallback
from talos.producer.TalosProducer import TalosProducer
from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.SimpleAbnormalCallback import SimpleAbnormalCallback
from talos.thrift.auth.ttypes import Credential
from talos.thrift.auth.ttypes import UserType
from talos.thrift.message.ttypes import Message
from atomic import AtomicLong
import logging
import time
import traceback

logger = logging.getLogger("TalosProducerDemo")
successPutNumber = AtomicLong(0)
talosProducer = TalosProducer


# callback for producer success/fail to put message
class MyMessageCallback(UserMessageCallback):
	# count when success

	def on_success(self, userMessageResult=None):
		successPutNumber.get_and_set(successPutNumber.value
									 + len(userMessageResult.get_message_list()))
		count = successPutNumber.value

		for message in userMessageResult.get_message_list():
			logger.info("success to put message: " + str(message.message))

		logger.info("success to put message: " + str(count) + " so far.")
		print("success to put message: " + str(count) + " so far.")

	# retry when failed
	def on_error(self, userMessageResult=None):
		try:
			for message in userMessageResult.get_message_list():
				logger.info("failed to put message: " + message + " we will retry to put it.")
			talosProducer.add_user_message(userMessageResult.get_message_list())
		except Exception as e:
			print(str(traceback.format_exc()))


class TalosProducerDemo:
	accessKey = "$yourAccessKey"
	accessSecret = "$yourSecretKey"
	topicName = "$yourTopicName"

	pro = dict()
	pro["galaxy.talos.service.endpoint"] = "$yourEndpoint"
	consumerConfig = TalosClientConfig
	credential = Credential
	toPutMsgNumber = 5

	def __init__(self):
		self.producerConfig = TalosClientConfig(self.pro)
		# credential
		self.credential = Credential(UserType.DEV_XIAOMI, self.accessKey, self.accessSecret)

	def start(self):
		# init producer
		talosProducer = TalosProducer(producerConfig=self.producerConfig,
									  credential=self.credential,
									  topicName=self.topicName,
									  topicAbnormalCallback=SimpleAbnormalCallback(),
									  userMessageCallback=MyMessageCallback())

		while True:
			messageList = []
			i = 0
			while i < self.toPutMsgNumber:
				message = Message(message="test message: this message is a text string.")
				messageList.append(message)
				i += 1
			talosProducer.add_user_message(messageList)
			time.sleep(10)

		# when call shutdown function,
		# the producer will wait all the messages in buffer to send to server
		# talosProducer.shutdown()


producerDemo = TalosProducerDemo()
producerDemo.start()




#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#

from talos.thrift.topic.ttypes import GetDescribeInfoRequest
from talos.thrift.topic.ttypes import TopicTalosResourceName
from talos.admin.TalosAdmin import TalosAdmin
from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.TalosClientFactory import TalosClientFactory
from talos.utils import Utils
from talos.utils.Utils import ReadWriteLock
from talos.consumer.MessageProcessorFactory import MessageProcessorFactory
from talos.consumer.MessageReaderFactory import MessageReaderFactory
from talos.thrift.consumer.ttypes import ConsumeUnit
from talos.thrift.consumer.ttypes import QueryWorkerRequest
from talos.thrift.consumer.ttypes import RenewResponse
from talos.thrift.consumer.ttypes import RenewRequest
from talos.thrift.consumer.ttypes import LockWorkerRequest
from talos.client.TalosErrors import InvalidArgumentError
from talos.consumer.PartitionFetcher import PartitionFetcher
from talos.thrift.common.ttypes import GalaxyTalosException
from threading import Timer
import logging
import random
import copy
import threading
import traceback


logger = logging.getLogger("TalosConsumer")


class WorkerPair(object):
    workerId = str
    hasPartitionNum = int

    def __init__(self, workerId=None, hasPartitionNum=None):
        self.workerId = workerId
        self.hasPartitionNum = hasPartitionNum

    def __lt__(self, other):
        return self.hasPartitionNum > other.hasPartitionNum

    def to_string(self):
        return "{'" + self.workerId + '\'' + ", " + str(self.hasPartitionNum) + '}'


class TalosConsumer(object):
    workerId = str
    random = random
    consumerGroup = str
    messageProcessorFactory = MessageProcessorFactory
    messageReaderFactory = MessageReaderFactory
    partitionFetcherMap = dict
    talosConsumerConfig = TalosClientConfig
    talosClientFactory = TalosClientFactory
    talosAdmin = TalosAdmin
    readWriteLock = ReadWriteLock()
    isActive = bool
    reBalanceExecutor = threading.Thread

    # 3 single scheduledExecutor respectively used for
    # a) checking partition number periodically
    # b) checking alive worker info periodically
    # c) renew worker heartbeat and serving partition locks periodically
    partitionScheduledExecutor = Timer
    workerScheduleExecutor = Timer
    renewScheduleExecutor = Timer

    # init by getting from rpc call as follows
    topicName = str
    partitionNumber = int
    topicTalosResourceName = TopicTalosResourceName
    workerInfoMap = dict
    partitionCheckPoint = dict

    def __init__(self, consumerGroup=None, consumerConfig=None,
                 credential=None, topicName=None, messageProcessorFactory=None,
                 clientPrefix=None, talosResourceName=None, workerId=None,
                 consumerClient=None, talosAdmin=None, fetcherMap=None,
                 partitionCheckPoint=None):
        if workerId:
            self.workerId = workerId
        else:
            self.workerId = Utils.generate_client_id(consumerConfig.get_client_ip(), clientPrefix)
        self.consumerGroup = consumerGroup
        self.talosConsumerConfig = consumerConfig
        self.topicName = topicName
        self.messageProcessorFactory = messageProcessorFactory
        self.messageReaderFactory = MessageReaderFactory()
        self.talosClientFactory = TalosClientFactory(consumerConfig, credential)
        if talosAdmin:
            self.talosAdmin = talosAdmin
        else:
            self.talosAdmin = TalosAdmin(self.talosClientFactory)
        if consumerClient:
            self.consumerClient = consumerClient
        else:
            self.consumerClient = self.talosClientFactory.new_consumer_client()
            self.lockClient = self.talosClientFactory.new_consumer_client()
            self.renewClient = self.talosClientFactory.new_consumer_client()
            self.queryClient = self.talosClientFactory.new_consumer_client()
        if talosResourceName:
            self.topicTalosResourceName = talosResourceName
        else:
            self.topicTalosResourceName = self.talosAdmin.get_describe_info(
                GetDescribeInfoRequest(self.topicName))
        if fetcherMap:
            self.partitionFetcherMap = fetcherMap
        else:
            self.partitionFetcherMap = dict()
        if partitionCheckPoint:
            self.partitionCheckPoint = partitionCheckPoint
        else:
            self.partitionCheckPoint = dict()
        self.isActive = True

        logger.info("The worker: " + self.workerId + " is initializing...")
        # check and get topic info such as partitionNumber
        self.check_and_get_topic_info(self.topicTalosResourceName)
        # register self workerId
        self.register_self()
        # get worker info
        self.get_worker_info()
        # do balance and init simple consumer
        self.make_balance()
        # start CheckPartitionTask/CheckWorkerInfoTask/RenewTask
        self.init_check_partition_task()
        self.init_check_worker_info_task()
        self.init_renew_task()

    #
    # Check Partition Task
    # if partition number change, invoke ReBalanceTask
    #
    def check_partition_task(self):
        try:
            request = GetDescribeInfoRequest(self.topicName)
            response = self.talosAdmin.get_describe_info(request)
        except Exception as throwable:
            logger.error("Exception in CheckPartitionTask: " + str(traceback.format_exc()))
            # if throwable instance of HBaseOperationFailed, just return
            # if throwable instance of TopicNotExist, cancel all reading task
            if Utils.is_topic_not_exist(throwable):
                self.cancel_all_consuming_task()
            self.partitionScheduledExecutor = Timer(interval=self.talosConsumerConfig
                                                    .get_check_partition_interval(),
                                                    function=self.check_partition_task)
            self.partitionScheduledExecutor.setName(
                "checkPartitionTask-" + self.topicName)
            self.partitionScheduledExecutor.start()
            return

        if not self.topicTalosResourceName == response.topicTalosResourceName:
            errMsg = "The topic: " + self.topicTalosResourceName.topicTalosResourceName \
                     + " not exist. It might have been deleted. " + \
                     "The getMessage threads will be cancel."
            logger.error(errMsg)
            self.cancel_all_consuming_task()
            self.partitionScheduledExecutor = Timer(interval=self.talosConsumerConfig
                                                    .get_check_partition_interval(),
                                                    function=self.check_partition_task)
            self.partitionScheduledExecutor.setName(
                "checkPartitionTask-" + self.topicName)
            self.partitionScheduledExecutor.start()
            return

        topicPartitionNum = response.partitionNumber
        if self.partitionNumber < topicPartitionNum:
            logger.info("partitionNumber changed from " + str(self.partitionNumber) + " to " +
                        str(topicPartitionNum) + ", execute a re-balance task.")
            # update partition number and call the re - balance
            self.set_partition_number(topicPartitionNum)
            # call the re - balance task
            self.reBalanceExecutor = threading.Thread(target=self.re_balance_task)
            self.reBalanceExecutor.setName("reBalanceTask-" + self.topicName)
            self.reBalanceExecutor.start()

        self.partitionScheduledExecutor = Timer(interval=self.talosConsumerConfig
                                                .get_check_partition_interval(),
                                                function=self.check_partition_task)
        self.partitionScheduledExecutor.setName("checkPartitionTask-" + self.topicName)
        self.partitionScheduledExecutor.start()

    #
    # Check Worker Info Task
    #
    # check alive worker number and get the worker serving map
    # 1) get the latest worker info and synchronized update the local workInfoMap
    # 2) invoke the ReBalanceTask every time
    #
    # Note:
    # a) current alive workers refer to scan 'consumerGroup+Topic+Worker'
    # b) all serving partitions got by the a)'s alive workers
    #
    # G+T+W    G+T+P
    # yes       no  -- normal, exist idle workers
    # no        yes -- abnormal, but ttl will fix it
    #
    def check_worker_info_task(self):
        try:
            self.get_worker_info()
        except Exception as throwble:
            logger.error("Get worker info error: " + str(traceback.format_exc()))
        # invoke the re-balance task every time
        self.reBalanceExecutor = threading.Thread(target=self.re_balance_task)
        self.reBalanceExecutor.setName("reBalanceTask-" + self.topicName)
        self.reBalanceExecutor.start()

        self.workerScheduleExecutor = Timer(interval=self.talosConsumerConfig
                                            .get_check_worker_info_interval(),
                                            function=self.check_worker_info_task)
        self.workerScheduleExecutor.setName("checkWorkerInfoTask-" + self.topicName)
        self.workerScheduleExecutor.start()

    #
    # Re - Balance Task
    #
    # This task just re - calculate the 'has' / 'max' / 'min' and try to steal / release
    # 'CheckPartitionTask' takes charge of updating partitionNumber
    # 'CheckWorkerInfoTask' takes charge of updating workerInfoMap
    #
    def re_balance_task(self):
        self.make_balance()

    #
    # ReNew Task(contains two sections per renew)
    #
    # Note: we make renew process outside rather than inner PartitionFetcher class
    # because:
    # 1) make the partitionFetcher heartbeat and worker heartbeat together
    # 2) renew all the serving partitions lock within one rpc process,
    # which prevent massive rpc request to server
    #
    # when get what to renew, we take 'partitionFetcherMap' as guideline
    #
    def renew_task(self):
        toRenewPartitionList = self.get_renew_partition_list()
        consumeUnit = ConsumeUnit(self.consumerGroup, self.topicTalosResourceName,
                                  toRenewPartitionList, self.workerId)
        renewRequest = RenewRequest(consumeUnit)
        renewResponse = RenewResponse

        # plus 1 to include the first renew operation
        maxRetry = self.talosConsumerConfig.get_renew_max_retry() + 1
        while maxRetry > 0:
            try:
                renewResponse = self.renewClient.renew(renewRequest)
            except Exception as e:
                if isinstance(e, GalaxyTalosException):
                    logger.error(e.errMsg)
                logger.error("Worker: " + self.workerId + " renew error: "
                             + str(traceback.format_exc()))
                continue
            maxRetry -= 1

            # 1) make heartbeat success and renew partitions success
            if renewResponse.heartbeatSuccess and len(renewResponse.failedPartitionList) == 0:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("The worker: " + self.workerId +
                                 " success heartbeat and renew partitions: " + str(toRenewPartitionList))
                break

        # 2) make heart beat failed, cancel all partitions
        # no need to renew anything, so block the renew thread and cancel all task
        if renewResponse and not renewResponse.heartbeatSuccess:
            logger.error("The worker: " + self.workerId + " failed to make heartbeat, cancel all consumer task")
            self.cancel_all_consuming_task()

        # 3) make heartbeat success but renew some partitions failed
        # stop read, commit offset, unlock for renew failed partitions
        # the release process is graceful, so may be a long time,
        # do not block the renew thread and switch thread to re-balance thread
        if renewResponse and len(renewResponse.failedPartitionList) > 0:
            failedRenewList = renewResponse.failedPartitionList
            logger.error("The worker: " + self.workerId + " failed to renew partitions: "
                         + str(failedRenewList))
            self.release_partition_lock(failedRenewList)

        self.renewScheduleExecutor = Timer(interval=self.talosConsumerConfig
                                           .get_renew_check_interval(),
                                           function=self.renew_task)
        self.renewScheduleExecutor.setName("renewTask-" + self.topicName)
        self.renewScheduleExecutor.start()

    def check_and_get_topic_info(self, topicTalosResourceName=None):
        topicName = Utils.get_topic_name_by_resource_name(
            str(topicTalosResourceName.topicTalosResourceName.topicTalosResourceName))
        response = self.talosAdmin.get_describe_info(GetDescribeInfoRequest(topicName))
        resourceName = topicTalosResourceName.topicTalosResourceName.topicTalosResourceName

        if not resourceName == response.topicTalosResourceName.topicTalosResourceName:
            logger.info("The consumer initialize failed by topic not found")
            raise InvalidArgumentError("The topic: " + resourceName +
                                       " not found")

        self.set_partition_number(response.partitionNumber)
        self.topicTalosResourceName = topicTalosResourceName.topicTalosResourceName
        logger.info("The worker: " + self.workerId + " check and get topic info done")

    def set_partition_number(self, partitionNumber=None):
        self.readWriteLock.write_acquire()
        self.partitionNumber = partitionNumber
        self.readWriteLock.write_release()

    def register_self(self):
        consumeUnit = ConsumeUnit(self.consumerGroup, self.topicTalosResourceName, [],
                                  self.workerId)
        lockWorkerRequest = LockWorkerRequest(consumeUnit)

        tryCount = self.talosConsumerConfig.get_self_register_max_retry() + 1
        while tryCount >= 0:
            try:
                lockWorkerResponse = self.lockClient.lock_worker(lockWorkerRequest)
            except Exception as e:
                if isinstance(e, GalaxyTalosException):
                    logger.error(e.errMsg)
                logger.error("The worker: " + self.workerId + "register self got error: "
                             + str(traceback.format_exc()))
                continue
            if lockWorkerResponse.registerSuccess:
                logger.info("The worker: " + self.workerId + " register self success")
                return
            logger.warn("The worker: " + self.workerId + " register self failed, make " +
                        str(tryCount) + " retry")
            tryCount -= 1

        logger.error("The worker: " + self.workerId + " register self failed")
        raise RuntimeError(self.workerId + " register self failed")

    def get_worker_info(self):
        queryWorkerRequest = QueryWorkerRequest(self.consumerGroup,
                                                self.topicTalosResourceName)
        queryWorkerResponse = self.queryClient.query_worker(queryWorkerRequest)

        # if queryWorkerInfoMap size equals 0,
        # it represents hbase failed error, do not update local map
        # because registration, the queryWorkerInfoMap size >= 1 at least
        # if queryWorkerInfoMap not contains self, it indicates renew failed,
        # do not update local map to prevent a bad re - balance
        if len(queryWorkerResponse.workerMap) == 0 or self.workerId not in \
                queryWorkerResponse.workerMap:
            return

        self.readWriteLock.write_acquire()
        self.workerInfoMap = queryWorkerResponse.workerMap
        self.readWriteLock.write_release()

    def calculate_target_list(self, copyPartitionNum=None, workerNumber=None,
                              targetList=None):
        if workerNumber == 1:
            # one worker serving all partitions
            targetList.append(copyPartitionNum)
        elif copyPartitionNum < workerNumber:
            # per worker per partition, the extra worker must be idle
            index = 0
            while index < copyPartitionNum:
                targetList.append(1)
                index += 1
        else:
            # calculate the target sequence
            sum = 0
            min = int(copyPartitionNum / workerNumber)
            remainder = int(copyPartitionNum % workerNumber)
            # add max by remainder
            index = 0
            while index < remainder:
                targetList.append(min + 1)
                sum += min + 1
                index += 1

            # add min by (workerNumber - remainder)
            index = 0
            while index < (workerNumber - remainder):
                targetList.append(min)
                sum += min
                index += 1
            assert sum == copyPartitionNum

        # sort target by descending
        targetList.sort(reverse=True)
        logger.info("worker: " + self.workerId + " calculate target partitions done: " +
                    str(targetList))

    def calculate_worker_pairs(self, copyWorkerMap=None, storedWorkerPairs=None):
        for key, value in copyWorkerMap.items():
            storedWorkerPairs.append(WorkerPair(key, len(value)))
        storedWorkerPairs.sort()
        logger.info("worker: " + self.workerId + " calculate sorted worker pairs: "
                    + str(storedWorkerPairs))

    def make_balance(self):
        #
        # When start make balance, we deep copy 'partitionNumber' and 'workerInfoMap'
        # to prevent both value appear inconsistent during the process makeBalance
        #
        copyPartitionNum = self.partitionNumber
        copyWorkerInfoMap = self.deep_copy_worker_info()

        #
        # if workerInfoMap not contains workerId, there must be error in renew task.
        # the renew task will cancel the consuming task and stop to read data,
        # so just return and do not care balance.
        #
        if not (self.workerId in copyWorkerInfoMap):
            logger.error("WorkerInfoMap not contains worker: " + self.workerId +
                         ". There may be some error for renew task.")
            return

        # calculate target and sorted worker pairs
        targetList = []
        sortedWorkerPairs = []
        self.calculate_target_list(copyPartitionNum, len(copyWorkerInfoMap), targetList)
        self.calculate_worker_pairs(copyWorkerInfoMap, sortedWorkerPairs)

        # judge stealing or release
        toStealList = []
        toReleaseList = []

        i = 0
        while i < len(sortedWorkerPairs):
            if sortedWorkerPairs[i].workerId == self.workerId:
                hasList = self.get_has_list()
                has = len(hasList)

                # workerNum > partitionNum, idle workers have no match target, do nothing
                if i >= len(targetList):
                    break
                target = targetList[i]
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Worker: " + self.workerId + " has: " + str(has) +
                                 " target: " + str(target))

                # a balanced state, do nothing
                if has == target:
                    return
                elif has > target:
                    # release partitions
                    toReleaseNum = has - target
                    while toReleaseNum > 0 and len(hasList) > 0:
                        toReleaseList.append(hasList.pop(0))
                        toReleaseNum -= 1
                else:
                    # stealing partitions
                    idlePartitions = self.get_idle_partitions()
                    if len(idlePartitions) > 0:
                        toStealNum = target - has
                        while toStealNum > 0 and len(idlePartitions) > 0:
                            randomIndex = random.randint(0, len(idlePartitions) - 1)
                            toStealList.append(idlePartitions.pop(randomIndex))
                            toStealNum -= 1
            i += 1

        # steal or release partition lock or reached a balance state
        assert not (len(toStealList) > 0 and len(toReleaseList) > 0)
        if len(toStealList) > 0:
            self.steal_partition_lock(toStealList)
        elif len(toReleaseList) > 0:
            self.release_partition_lock(toReleaseList)
        else:
            # do nothing when reach balance state
            logger.info("The worker: " + self.workerId + " have reached a balance state.")

    def steal_partition_lock(self, toStealList=None):
        logger.info("Worker: " + self.workerId + " try to steal " + str(len(toStealList)) +
                    " partition: " + str(toStealList))
        # try to lock and invoke serving partition PartitionFetcher to 'LOCKED' state
        self.readWriteLock.write_acquire()
        for partitionId in toStealList:
            if partitionId not in self.partitionFetcherMap:
                # Note 'partitionCheckPoint.get(partitionId)' may be null, it's ok
                partitionFetcher = PartitionFetcher(consumerGroup=self.consumerGroup,
                                                    topicName=self.topicName,
                                                    topicTalosResourceName=self.topicTalosResourceName,
                                                    partitionId=partitionId,
                                                    talosConsumerConfig=self.talosConsumerConfig,
                                                    workerId=self.workerId,
                                                    consumerClient=self.talosClientFactory.new_consumer_client(),
                                                    messageClient=self.talosClientFactory.new_message_client(),
                                                    queryPartitionCleint=self.talosClientFactory.new_message_client(),
                                                    messageProcessor=self.messageProcessorFactory.create_processor(),
                                                    messageReader=self.messageReaderFactory.create_message_reader(self.talosConsumerConfig),
                                                    outerCheckPoint=self.partitionCheckPoint.get(partitionId))
                self.partitionFetcherMap[partitionId] = partitionFetcher
            self.partitionFetcherMap.get(partitionId).lock()
        self.readWriteLock.write_release()

    def release_partition_lock(self, toReleaseList=None):
        logger.info("Worker: " + self.workerId + " try to release " + str(len(toReleaseList))
                    + " partition: " + str(toReleaseList))
        # stop read, commit offset, unlock the partition async
        for partitionId in toReleaseList:
            assert partitionId in self.partitionFetcherMap
            self.partitionFetcherMap.get(partitionId).unlock()
            del self.partitionFetcherMap[partitionId]

    def get_has_list(self):
        hasList = []
        self.readWriteLock.write_acquire()
        for key, value in self.partitionFetcherMap.items():
            if value.is_serving():
                hasList.append(key)
        self.readWriteLock.write_release()
        return hasList

    def get_idle_partitions(self):
        self.readWriteLock.read_acquire()
        assert self.partitionNumber > 0
        idlePartitions = []
        i = 0
        while i < self.partitionNumber:
            idlePartitions.append(i)
            i += 1
        for valueList in self.workerInfoMap.values():
            for partitionId in valueList:
                idlePartitions.remove(partitionId)
        self.readWriteLock.read_release()
        return idlePartitions

    def shut_down_fetcher(self):
        for i in self.partitionFetcherMap:
            self.partitionFetcherMap.get(i).shutdown()

    def deep_copy_worker_info(self):
        self.readWriteLock.read_acquire()
        copyMap = copy.deepcopy(self.workerInfoMap)
        self.readWriteLock.read_release()
        return copyMap

    def get_worker_id(self):
        return self.workerId

    def get_topic_talos_resource_name(self):
        return self.topicTalosResourceName

    def cancel_all_consuming_task(self):
        self.release_partition_lock(self.get_has_list())

    def get_renew_partition_list(self):
        toRenewList = []
        self.readWriteLock.read_acquire()
        for key, value in self.partitionFetcherMap.items():
            if value.is_holding_lock():
                toRenewList.append(key)
        self.readWriteLock.read_release()
        return toRenewList

    def init_check_partition_task(self):
        # check and update partition number every 1 minutes delay by default
        self.partitionScheduledExecutor = Timer(interval=self.talosConsumerConfig
                                                .get_check_partition_interval(),
                                                function=self.check_partition_task)
        self.partitionScheduledExecutor.setName("checkPartitionTask-" + self.topicName)
        self.partitionScheduledExecutor.start()

    def init_check_worker_info_task(self):
        self.workerScheduleExecutor = Timer(interval=self.talosConsumerConfig
                                            .get_check_worker_info_interval(),
                                            function=self.check_worker_info_task)
        self.workerScheduleExecutor.setName("checkWorkerInfoTask-" + self.topicName)
        self.workerScheduleExecutor.start()

    def init_renew_task(self):
        self.renewScheduleExecutor = Timer(interval=self.talosConsumerConfig
                                           .get_renew_check_interval(),
                                           function=self.renew_task)
        self.renewScheduleExecutor.setName("renewTask-" + self.topicName)
        self.renewScheduleExecutor.start()

    #
    # First stop CheckPartition and CheckWorkerInfo Task and then shutdown all fetcher,
    # that avoid thread exit failed caused by rebalance partition when shutDownAllFetcher
    #
    def shutdown(self):
        if not self.isActive:
            logger.info("Worker: " + str(self.workerId) + " is already shutdown, "
                        + "don't do it again.")
            return
        logger.info("Worker: " + self.workerId + " is shutting down...")
        self.partitionScheduledExecutor.cancel()
        self.workerScheduleExecutor.cancel()
        self.shut_down_fetcher()
        self.renewScheduleExecutor.cancel()
        self.isActive = False
        logger.info("Worker: " + self.workerId + " shutdown.")



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
#
from talos.thrift.auth.ttypes import Credential
from talos.client.TalosClientFactory import MessageClient
from talos.client.TalosClientFactory import ConsumerClient
from talos.client.TalosClientFactory import TopicClient
from talos.client.TalosClientFactory import QuotaClient
import logging


class TalosAdmin(object):
    logger = logging.getLogger("TalosAdmin")
    topicClient = TopicClient
    messageClient = MessageClient
    consumerClient = ConsumerClient
    quotaClient = QuotaClient
    credential = Credential()

    def __init__(self, talosClientFactory=None):
        self.topicClient = talosClientFactory.new_topic_client()

    # topicAttribute for partitionNumber required

    def get_describe_info(self, request=None):
        return self.topicClient.get_describe_info(request)

    def get_topic_attribute(self, request=None):
        return self.topicClient.get_topic_attribute(request)

    def create_topic(self, request=None):
        return self.topicClient.create_topic(request)

    def delete_topic(self, request=None):
        return self.topicClient.delete_topic(request)

    def change_topic_attribute(self, request=None):
        return self.topicClient.change_topic_attribute(request)

    def list_topics(self):
        return self.topicClient.list_topics()

    def list_permissions(self):
        return self.topicClient.list_permission()

    def set_permission(self, request=None):
        return self.topicClient.set_permission(request)

    def revoke_permission(self, request=None):
        return self.topicClient.revoke_permission(request)

    def get_schedule_info(self, request=None):
        return self.messageClient.get_schedule_info(request)

    def apply_quota(self, request=None):
        return self.quotaClient.apply_quota(request)

    def revoke_quota(self, request=None):
        return self.quotaClient.revoke_quota(request)

    def list_quota(self):
        return self.quotaClient.list_quota

    def get_topic_offset(self, request=None):
        return self.messageClient.get_topic_offset(request)

    def get_partition_offset(self, request=None):
        return self.messageClient.get_partition_offset(request)





#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

import abc


class TopicAbnormalCallback:
	#
	# User implement this method to process topic abnormal status such as 'TopicNotExist'
	#
	@abc.abstractmethod
	def abnormal_handler(self, topicTalosResourceName=None, throwable=None):
		pass



#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 
 
from abc import abstractmethod


class Partitioner:

    @abstractmethod
    def partition(self, partitionKey=None, partitionNum=None):
        pass



import unittest

from talos.thrift.message.ttypes import Message
from talos.producer.UserMessage import UserMessage


class test_UserMessage(unittest.TestCase):
	partitionKey = "123456qwerty"
	sequenceNumber = "654321"
	data = "hello"

	def setUp(self):
		self.message = Message(message=self.data, partitionKey=self.partitionKey)
		self.userMessage = UserMessage(self.message)

	def tearDown(self):
		pass

	def test_get_message(self):
		self.assertEqual(self.message, self.userMessage.get_message())

	def test_get_timestamp(self):
		self.assertIsNotNone(self.userMessage.get_timestamp())

	def test_get_message_size(self):
		self.assertEqual(len(self.data), self.userMessage.get_message_size())

		self.message.sequenceNumber = self.sequenceNumber
		self.userMessage = UserMessage(self.message)
		self.assertEqual(len(self.data) + len(self.sequenceNumber), self.userMessage.get_message_size())


if __name__ == '__main__':
	unittest.main()



# -*- coding:utf8 -*-
#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 

from talos.client.TalosClientConfig import TalosClientConfig
from talos.client.compression.Compression import Compression
from talos.client.TalosClientFactory import TalosClientFactory
from talos.thrift.topic.ttypes import TopicAndPartition
from talos.thrift.topic.TopicService import GetDescribeInfoRequest
from talos.thrift.message import MessageService
from talos.thrift.message.ttypes import MessageType
from talos.thrift.message.ttypes import PutMessageRequest
from atomic import AtomicLong
from talos.utils import Utils
import logging
import traceback


class SimpleProducer(object):
    logger = logging.getLogger("SimpleConsumer")

    producerConfig = TalosClientConfig
    topicAndPartition = TopicAndPartition
    messageClient = MessageService.Iface
    requestId = AtomicLong
    clientId = str
    isActive = bool

    def __init__(self, producerConfig=None, topicName=None, topicAndPartition=None,
                 partitionId=None, credential=None, talosClientFactory=None,
                 messageClient=None, clientId=None, requestId=None):
        if talosClientFactory:
            self.talosClientFactory = talosClientFactory
        elif credential:
            self.talosClientFactory = TalosClientFactory(producerConfig, credential)
        if topicName:
            Utils.check_topic_name(topicName)
            self.get_topic_info(self.talosClientFactory.new_topic_client(), topicName,
                                partitionId)
        else:
            self.topicAndPartition = topicAndPartition
        self.producerConfig = producerConfig
        if messageClient:
            self.messageClient = messageClient
        else:
            self.messageClient = self.talosClientFactory.new_message_client()
        if clientId:
            self.clientId = clientId
        else:
            self.clientId = Utils.generate_client_id('SimpleProducer', '')
        if requestId:
            self.requestId = requestId
        else:
            self.requestId = AtomicLong(1)
        self.isActive = True

    def get_topic_info(self, topicClient=None, topicName=None, partitionId=None):
        response = topicClient.get_describe_info(GetDescribeInfoRequest(topicName))
        self.topicAndPartition = TopicAndPartition(topicName=topicName,
                                                   topicTalosResourceName=response.topicTalosResourceName,
                                                   partitionId=partitionId)

    def put_message(self, msgList=None):
        if (not msgList) or len(msgList) == 0:
            return True

        try:
            self.put_message_list(msgList)
            return True
        except Exception as e:
            self.logger.error("putMessage error， please try to put again"
                              + str(traceback.format_exc()))

        return False

    def put_message_list(self, msgList=None):
        if (not msgList) or len(msgList) == 0:
            return

        # check data validity
        for message in msgList:
            # set timestamp and messageType if not set
            Utils.update_message(message, MessageType.BINARY)

        # check data validity
        Utils.check_message_list_validity(msgList)

        self.do_put(msgList)

    def do_put(self, msgList=None):
        messageBlock = self._compress_message_list(msgList)
        messageBlockList = [messageBlock]

        requestSequenceId = Utils.generate_request_sequence_id(self.clientId,
                                                               self.requestId)
        putMessageRequest = PutMessageRequest(self.topicAndPartition, messageBlockList,
                                              len(msgList), requestSequenceId)
        try:
            putMessageResponse = self.messageClient.put_message(putMessageRequest)
        except Exception as e:
            self.logger.error("put message request failed." + str(traceback.format_exc()))
            raise e

    def _compress_message_list(self, msgList=None):
        return Compression().compress(msgList, self.producerConfig.get_compression_type())







#
# Copyright 2020, Xiaomi.
# All rights reserved.
# Author: huyumei@xiaomi.com
# 
 
from talos.client.serialization.MessageSerializer import MessageSerializer
from talos.thrift.message.ttypes import Message
from talos.thrift.message.ttypes import MessageType
from talos.client.serialization import MessageVersion
from struct import unpack, pack_into
from talos.utils.Utils import SerializeFormat
from talos.utils import Utils
import logging


class MessageSerializerV2(MessageSerializer):

    logger = logging.getLogger("MessageSerializerV2")

    __CREATE_TIMESTAMP_BYTES_V2 = 8
    __MESSAGE_TYPE_BYTES_V2 = 2
    __SEQUENCE_NUMBER_LENGTH_BYTES_V2 = 2
    __VERSION_NUMBER_LENGTH_V2 = 4
    __MESSAGE_DATA_LENGTH_BYTES_V2 = 4
    __MESSAGE_HEADER_BYTES_V2 = __VERSION_NUMBER_LENGTH_V2 + __MESSAGE_TYPE_BYTES_V2 + \
                                __CREATE_TIMESTAMP_BYTES_V2 + \
                                __SEQUENCE_NUMBER_LENGTH_BYTES_V2 + \
                                __MESSAGE_DATA_LENGTH_BYTES_V2

    # TalosProducer serialize
    def serialize(self, msg=None, buf=None):
        # write version number
        self.write_message_version(MessageVersion.MessageVersion(2), buf)

        # write create timestamp
        timestampBuffer = bytearray(self.__CREATE_TIMESTAMP_BYTES_V2)
        if msg.createTimestamp:
            pack_into(SerializeFormat.format_i64, timestampBuffer, 0, msg.createTimestamp)
        else:
            pack_into(SerializeFormat.format_i64, timestampBuffer, 0,
                      Utils.current_time_mills())
        buf.write(timestampBuffer)

        # write message type
        if not msg.messageType:
            raise RuntimeError("message must set messageType")
        else:
            messageTypeBuffer = bytearray(self.__MESSAGE_TYPE_BYTES_V2)
            pack_into(SerializeFormat.format_i16, messageTypeBuffer, 0, msg.messageType)
            buf.write(messageTypeBuffer)

        # write sequence number
        sequenceNumberBuffer = bytearray(self.__SEQUENCE_NUMBER_LENGTH_BYTES_V2)
        if msg.sequenceNumber:
            pack_into(SerializeFormat.format_i16, sequenceNumberBuffer, 0,
                      len(msg.sequenceNumber))
            buf.write(sequenceNumberBuffer)
            buf.write(msg.sequenceNumber)
        else:
            pack_into(SerializeFormat.format_i16,
                                             sequenceNumberBuffer, 0, 0)
            buf.write(sequenceNumberBuffer)

        # write message size
        sizeBuffer = bytearray(self.__MESSAGE_DATA_LENGTH_BYTES_V2)
        pack_into(SerializeFormat.format_i32, sizeBuffer, 0, len(msg.message))
        buf.write(sizeBuffer)

        # write message
        buf.write(msg.message)
        return

    # TalosConsumer deserialize
    def deserialize(self, header=None, buf=None):
        msg = Message
        # read create timestamp
        timestampBuffer = bytearray(self.__CREATE_TIMESTAMP_BYTES_V2)
        try:
            if buf.readinto(timestampBuffer) != self.__CREATE_TIMESTAMP_BYTES_V2:
                self.logger.error("deserialize create timestamp error!")
                return
        except Exception as e:
            self.logger.error("deserialize create timestamp error!" + e.message)
            raise e
        else:
            createTimestamp = int(unpack(SerializeFormat.format_i64, timestampBuffer)[0])
            msg.createTimestamp = createTimestamp

        # read message type
        messageTypeBuffer = bytearray(self.__MESSAGE_TYPE_BYTES_V2)
        try:
            if buf.readinto(messageTypeBuffer) != self.__MESSAGE_TYPE_BYTES_V2:
                self.logger.error("deserialize message type error!")
                return
        except Exception as e:
            self.logger.error("deserialize message type error!" + e.message)
            raise e
        else:
            messageType = int(unpack(SerializeFormat.format_i16, messageTypeBuffer)[0])
            if messageType == MessageType.UTF8:
                msg.messageType = MessageType.UTF8
            elif messageType == MessageType.BINARY:
                msg.messageType = MessageType.BINARY
            elif messageType == MessageType.AVRO:
                msg.messageType = MessageType.AVRO
            elif messageType == MessageType.PROTOBUF:
                msg.messageType = MessageType.PROTOBUF
            else:
                msg.messageType = MessageType.THRIFT

        # read sequence number
        sequenceNumLenBuffer = bytearray(self.__SEQUENCE_NUMBER_LENGTH_BYTES_V2)
        try:
            if buf.readinto(sequenceNumLenBuffer) != self.__SEQUENCE_NUMBER_LENGTH_BYTES_V2:
                self.logger.error("deserialize sequence number length error!")
                return
        except Exception as e:
            self.logger.error("deserialize sequence number length error!" + e.message)
            raise e
        else:
            sequenceNumLen = int(unpack(SerializeFormat.format_i16, sequenceNumLenBuffer)[0])

        if not sequenceNumLen == 0:
            sequenceNumberBuffer = bytearray(sequenceNumLen)
            try:
                if buf.readinto(sequenceNumberBuffer) != sequenceNumLen:
                    self.logger.error("deserialize sequence number error!")
                    return
            except Exception as e:
                self.logger.error("deserialize sequence number error!" + e.message)
                raise e
            else:
                sequenceNumber = str(sequenceNumberBuffer)
                msg.sequenceNumber = sequenceNumber

        # read message size
        messageSizeBuffer = bytearray(self.__MESSAGE_DATA_LENGTH_BYTES_V2)
        try:
            if buf.readinto(messageSizeBuffer) != self.__MESSAGE_DATA_LENGTH_BYTES_V2:
                self.logger.error("deserialize message size error!")
                return
        except Exception as e:
            self.logger.error("deserialize message size error!" + e.message)
            raise e
        else:
            messageSize = int(unpack(SerializeFormat.format_i32, messageSizeBuffer)[0])

        # read message
        messageDataBuffer = bytearray(messageSize)
        try:
            if buf.readinto(messageDataBuffer) != messageSize:
                self.logger.error("deserialize message error!")
                return
        except Exception as e:
            self.logger.error("deserialize message error!" + e.message)
            raise e
        else:
            msg.message = messageDataBuffer

        return msg

    def get_message_size(self, msg=None):
        size = self.__MESSAGE_HEADER_BYTES_V2
        if msg.sequenceNumber:
            size += len(msg.sequenceNumber)
        size += len(msg.message)
        return size


